/*
 * class_model_tau.c
 * 2015/10/25 ymaslenko $
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "gsl_sf_erf.h"
#include "gsl_linalg.h"
#include "gsl_permutation.h"
#include "gsl_matrix_double.h"
#include "gsl_blas.h"
#include "e_malloc.h"
#include "mprintf.h"

#include "amino_a.h"
#include "fio.h"
#include "bad_angle.h"
#include "class_model_tau.h"


/* ---------------- compare -----------------------------------
 * compares to size_t values ( used for qsort() )
 */
static int
compare(const void *a, const void *b) {
    return ((int) (*(size_t *) a - *(size_t *) b));
}


/* ---------------- clssfcn_destroy ----------------------------
 * frees up the memory used by some classification
 */
/*
void
clssfcn_destroy(struct clssfcn7 *c) {
    size_t i, j;
    if (c != NULL) {

        if (c->classmodel7 != NULL) {
            for (i = 0; i < c->n_class; i++)
                free_if_not_null(c->classmodel7[i]);
            free(c->classmodel7);
        }

        if (c->cov_matrix7 != NULL) {
            for (i = 0; i < c->n_class; i++)
                free_if_not_null(c->cov_matrix7[i]);
            free(c->cov_matrix7);
        }

        if (c->params7 != NULL) {
            for (i = 0; i < c->n_class; i++) {
                if (c->params7[i] != NULL) {
                    for (j = 0; j < c->dim; j++)
                        free_if_not_null(c->params7[i][j]);
                    free(c->params7[i]);
                }
            }
            free(c->params7);
        }

        free_if_not_null(c->class_weight7);

        free(c);
    }
}
*/

static void
scanf_error(struct clssfcn7 *cmodel, size_t i, size_t *attr_idx,
            const char *influence_report_filename) {
    const char this_sub[] = "scanf_error";

    long i_ = 0;

    err_printf(this_sub, "unexpected EOF reached in file %s!\n",
               influence_report_filename);
    for (i_ = i; i_ >= 0; i_--) {
        free_if_not_null(cmodel->classmodel7[i_]);
        free_if_not_null(cmodel->params7[i_]);
    }
    free_if_not_null(cmodel->classmodel7);
    free_if_not_null(cmodel->params7);
    free_if_not_null(attr_idx);
    free_if_not_null(cmodel->class_weight7);
    free_if_not_null(cmodel);
}

/* ---------------- tau_get_classfcn -------------------------------
 * Reads an influence report (only report_mode = "data" generated by
 * AutoClass-C
 * Do not forget to give an absolute error value from your original data
 * It is used for the integration in computeMembership().
 */
#define DUMMY_S_SIZE 12
#define DUMMY_STR_SIZE 80

struct clssfcn7 *
tau_get_clssfcn(const char *influence_report_filename, const float abs_error_dist, const float abs_error_tau) {

    const char this_sub[] = "tau_get_clssfcn";
    /*mprintf("%s\n", this_sub);*/
    size_t dummy_i = 0;
    float dummy_f = 0.0;
    char dummy_c;
    char *dummy_s = E_MALLOC(DUMMY_S_SIZE * sizeof(char));
    char *dummy_str = E_CALLOC(DUMMY_STR_SIZE, sizeof(char));
    float mean = 0, dev = 0;
    size_t corr_idx = 0;
    size_t i = 0, i_ = 0, j = 0, j_ = 0, j__ = 0, k = 0, l = 0;
    fpos_t pos, pos_, pos__;


    FILE *influence_report_file = NULL;

    struct clssfcn7 *cmodel = E_CALLOC(1, sizeof(struct clssfcn7));

    size_t n_cases = 0; /* (used to calculate a normalized class weight) */
    size_t *attr_idx = NULL; /* actual array indices for storing in cmodel */
    int *use_idx = NULL; /* which part of the correlation matrix is found? */

    memset(dummy_s, 0, DUMMY_S_SIZE);
    memset(dummy_str, 0, DUMMY_STR_SIZE);
    if ((influence_report_file = mfopen(influence_report_filename, "rb",
                                        this_sub)) == NULL) {
        return NULL;
    }
    cmodel->abs_error_tau = fabsf(abs_error_tau);
    cmodel->abs_error_dist = fabsf(abs_error_dist);

    /* looks for the number of cases */
    while (fscanf(influence_report_file,
                  " DATA_CLSF_HEADER %*s %*s %*s %*s %u",
                  (unsigned int *) &n_cases) != 1) {
        if (feof(influence_report_file)) {
            err_printf(this_sub,
                       "unexpected EOF reached in file %s!\n",
                       influence_report_filename);
            free_if_not_null(cmodel);
            return NULL;
        }
    }
    /* looks for the number of classes */
    while (fscanf(influence_report_file,
                  "%*s DATA_POP_CLASSES %*s %*s %u",
                  (unsigned int *) &cmodel->n_class) != 1) {
        if (feof(influence_report_file)) {
            err_printf(this_sub,
                       "unexpected EOF reached in file %s!\n",
                       influence_report_filename);
            free_if_not_null(cmodel);
            return NULL;
        }
    }
    if (cmodel->n_class < 1) {
        err_printf(this_sub,
                   "not even a single class found in file %s!\n",
                   influence_report_filename);
        free_if_not_null(cmodel);
        return NULL;
    }


    cmodel->class_weight7 = E_CALLOC(cmodel->n_class, sizeof(float));
    cmodel->cov_matrix7 = E_CALLOC(cmodel->n_class, sizeof(float *));

    /* looks for the beginning of attribute list */
    while (fscanf(influence_report_file,
                  "%*s DATA_NORM_INF_VALS "
                          "%*s %*s %*s %*s %*s %*s %*s %*s %*s %*s %*s "
                          "num description %s",
                  dummy_str) != 1 && strcmp(dummy_str, "I-*k")) {
        if (feof(influence_report_file)) {
            err_printf(this_sub,
                       "unexpected EOF reached in file %s!\n",
                       influence_report_filename);
            free_if_not_null(cmodel->class_weight7);
            free_if_not_null(cmodel);
            return NULL;
        }
    }
    /* Counts the number of actually used attributes, that is the dimension */
    if (fgetpos(influence_report_file, &pos) != 0) {
        err_printf(this_sub, "Fail in fsetpos in influence file\n");
    }

    while (fscanf(influence_report_file, "%*u %12c %f", dummy_s, &dummy_f) == 2) {
        cmodel->dim++;
    }

    if (fsetpos(influence_report_file, &pos) == -1) {
        err_printf(this_sub, "Fail in fsetpos in influence file\n");
        fclose(influence_report_file);
        free(cmodel);
        return NULL;
    }

    use_idx = E_CALLOC(cmodel->dim, sizeof(int));
    attr_idx = E_CALLOC(cmodel->dim, sizeof(size_t));

    /* Reads the list of attributes and sorts it to get a mapping between
     * attribute indices and the array to store them */
    while (fscanf(influence_report_file, "%u %12c %f",
                  (unsigned int *) &j, dummy_s, &dummy_f) == 3) {
        if (dummy_s[0] == 'c' || dummy_s[0] == 's' || dummy_s[0] == 'd') {
            attr_idx[i++] = j;
            fgetpos(influence_report_file, &pos_);
        }
    }

    fsetpos(influence_report_file, &pos_);
    qsort(attr_idx, cmodel->dim, sizeof(attr_idx[0]), compare);

    cmodel->params7 = E_CALLOC(cmodel->n_class, sizeof(double **));
    cmodel->classmodel7 = E_CALLOC(cmodel->n_class, sizeof(size_t * ));

    /* Reads the descriptions for all classes */
    for (i = 0; i < cmodel->n_class; i++) {
        cmodel->params7[i] = E_CALLOC(cmodel->dim, sizeof(double *));
        cmodel->cov_matrix7[i] = E_CALLOC(cmodel->dim * cmodel->dim, sizeof(double));

        for (j = 0; j < cmodel->dim; j++) {
            cmodel->params7[i][j] = E_CALLOC(3, sizeof(double));
        }
        cmodel->classmodel7[i] = E_CALLOC(cmodel->dim, sizeof(size_t));

        /* finds the beginning of a class discription and stores its weight */
        while (fscanf(influence_report_file,
                      "%s DATA_CLASS %*u CLASS %*u - weight %f",
                      dummy_str, &cmodel->class_weight7[i]) != 2) { //dummy_str== "-----"
            if (feof(influence_report_file)) {
                scanf_error(cmodel, i, attr_idx, influence_report_filename);
                return NULL;
            }
        }
        cmodel->class_weight7[i] /= n_cases;

        /* Looks for the beginning of the description list */
        while ((fscanf(influence_report_file,
                       "%*s REAL ATTRIBUTE (t = R) |Mean-jk - "
                               "numb t mtt description "
                               "I-jk Mean StDev Mean-*k|/ Mean StDev "
                               "t a -jk -jk StDev-jk -*k -*%c",
                       &dummy_c) != 1) || (dummy_c != 'k')) {
            if (feof(influence_report_file)) {
                scanf_error(cmodel, i, attr_idx, influence_report_filename);
                return NULL;
            }
        }

        i_ = cmodel->dim;

        while (i_) {
            if (fscanf(influence_report_file,
                       "%lu %lu %*s %s %12c"
                               " %*f %*c %e  %e%*c %*e %*c%*e %*e%*c",
                       &corr_idx, &j,
                       dummy_str, dummy_s, &mean, &dev) != 6) {
                scanf_error(cmodel, i, attr_idx, influence_report_filename);
                return NULL;
            }
            if (dummy_s[0] == 'c' || dummy_s[0] == 's' || dummy_s[0] == 'd') {

                /* Maps the attribute index to the storage index */
                j = (size_t *) bsearch(&j, attr_idx, cmodel->dim, sizeof(attr_idx[0]), compare) - attr_idx;

                if (dummy_s[0] == 'd') { /* distance is zero bounded */
                    cmodel->params7[i][j][0] = mean;
                    cmodel->params7[i][j][1] = dev;
                } else { /* cos and sin are 0 bounded */
                    cmodel->params7[i][j][0] = mean;
                    cmodel->params7[i][j][1] = dev;
                }
                /*  decides which model this dimension uses */
                if (strcmp(dummy_str, "MNcn") == 0) {
                    cmodel->classmodel7[i][j] = MULTI_NORMAL;
                    cmodel->params7[i][j][2] = corr_idx;
                }
                else if (strcmp(dummy_str, "SNcn") == 0) {
                    cmodel->classmodel7[i][j] = SINGLE_NORMAL;
                }
                else {
                    cmodel->classmodel7[i][j] = UNKNOWN;
                }
                i_--;
            }

            else {
                scanf_error(cmodel, i, attr_idx, influence_report_filename);
                return NULL;
            }
        }

        /* Tries to read the correlation matrices */
        fgetpos(influence_report_file, &pos);
        if (fscanf(influence_report_file, "%s", dummy_str) != 1) {
            scanf_error(cmodel, i, attr_idx, influence_report_filename);
            return NULL;
        }
        while (strcmp("DATA_CORR_MATRIX", dummy_str)) {
            if (feof(influence_report_file)) {
                scanf_error(cmodel, i, attr_idx, influence_report_filename);
                return NULL;
            }
            fgetpos(influence_report_file, &pos);
            if (fscanf(influence_report_file, "%s", dummy_str) != 1) {
                scanf_error(cmodel, i, attr_idx, influence_report_filename);
                return NULL;
            }
        }

        /* set the pos to the beginning of the "DATA_CORR_MATRIX" */
        fsetpos(influence_report_file, &pos);
        while (fscanf(influence_report_file, "%s ", dummy_str) == 1
               && strcmp(dummy_str, "DATA_CORR_MATRIX") == 0) {
            fgetpos(influence_report_file, &pos);

            while (fscanf(influence_report_file, "%*s %lu", &dummy_i) != 1) {
                fgetpos(influence_report_file, &pos);
                pos__ = pos_ = pos;
            }

            /* maps the attribute index to the storage index */
            j = (size_t *) bsearch(&dummy_i, attr_idx,
                                   cmodel->dim, sizeof(attr_idx[0]), compare) - attr_idx;
            use_idx[j] = 1;

            j__ = j_ = dummy_i;
            while (fscanf(influence_report_file, "%lu", &dummy_i) == 1) {
                j__ = j_;     /* saves previously read numbers to have the */
                j_ = dummy_i; /* right ones by hand */
                /* maps the attribute index to the storage index */
                j = (size_t *) bsearch(&j__, attr_idx,
                                       cmodel->dim, sizeof(attr_idx[0]), compare) - attr_idx;
                use_idx[j] = 1;
                pos__ = pos_;            /* saves previous file positions to */
                pos_ = pos;              /* to jump back later */
                fgetpos(influence_report_file, &pos);
            }
            /* Maps the attribute index to the storage index */
            j = (size_t *) bsearch(&j__, attr_idx,
                                   cmodel->dim,
                                   sizeof(attr_idx[0]), compare)
                - (size_t *) attr_idx;
            use_idx[j] = 1;
            pos = pos__;

            /* Reads the correlation values for the found attributes */
            fsetpos(influence_report_file, &pos);
            for (k = 0; k < cmodel->dim; k++) {
                if (use_idx[k] == 0)
                    continue;
                if (fscanf(influence_report_file, "%u", (unsigned int *) &dummy_i) != 1) {
                    scanf_error(cmodel, i, attr_idx, influence_report_filename);
                    return NULL;
                }

                for (l = 0; l < cmodel->dim; l++) {
                    if (use_idx[l] == 0)
                        continue;
                    if (fscanf(influence_report_file, "%lf",
                               &cmodel->cov_matrix7[i][(k * cmodel->dim) + l]) != 1) {
                        scanf_error(cmodel, i, attr_idx, influence_report_filename);
                        return NULL;
                    }
                    cmodel->cov_matrix7[i][(k * cmodel->dim) + l] *=
                            cmodel->params7[i][k][1] * cmodel->params7[i][l][1];
                }
            }
            for (k = 0; k < cmodel->dim; k++)
                use_idx[k] = 0;
        }
        fsetpos(influence_report_file, &pos);
    }

    free_if_not_null(dummy_s);
    free_if_not_null(dummy_str);
    free_if_not_null(use_idx);
    free_if_not_null(attr_idx);
    fclose(influence_report_file);
//     mprintf("leaving %s...\n", this_sub);
    return cmodel;
}

/* ---------------------- valid_distance_or  ------------------------
 * help_function for computeMembershipStrct
 * check the distance value
*/
double
valid_distance_or(float distance, float default_) {
    return distance == BAD_DISTANCE_FLOAT ? default_ : distance;
}

/* ---------------------- valid_in_range_or  ------------------------
 * help_function for computeMembershipStrct
 * check the cos/sin value
*/
double
valid_in_range_or(float value, float min, float max, float default_) {
    return (value > min && value < max) ? value : default_;
}

/* ------------------------ calc_dist ------------------------
 * help_function for computeMembershipStrct
 * check the value of the test_array on the position j
 * return 0 if value is corrupt
 * and calculate distance for accurate numbers
 * flag is used for angles
*/
static double
distance_to_mean(double test_vec_value, double mean, bool angle_normalize) {
    double dist = 0.0;
    if (test_vec_value != 0.0)
        dist = test_vec_value - mean;
    if (angle_normalize) {
        if (dist > M_PI) {
            dist -= 2 * M_PI;
        } else if (dist < -M_PI) {
            dist += 2 * M_PI;
        }
    }

    return dist;
}

/* ---------------- tau_computeMemebershipStrct -------------------------
 * For a case descriptor test_vec (of dimension cmodel->dim)
 * compute the memberships to each class in cmodel.
 * Returns the vector of memberships (length is cmodel->n_class)
 * or NULL if an unsupported class is found in cmodel.
 */
/* 9 Nov 2015 YMaslenko
 * This function was broken into two parts:
 * computeMembership and computeMemberStrct
 * It shouldn't affect any routine which calls computeMembership
 */
/* --------------- tau_computeMembershipStrct ----------------------
 * Core of the original function computeMembershipStrct
 * mship should be initialized, normally to class_weight7 or 1
 */
#define square(x)( (x) * (x) )

float *
tau_computeMembershipStrct(float *mship, const float *test_vec,
                           const struct clssfcn7 *cmodel) {

    const char this_sub[] = "tau_computeMembershipStrct";
    int class_idx, attr_idx;

    if (mship == NULL) {
        err_printf(this_sub, "mship is null\n");
        return NULL;
    }
    if (test_vec == NULL) {
        err_printf(this_sub, "test_vec is NULL\n");
        return NULL;
    }

    const int MEAN = 0;
    const int STDDEV = 1;
    double *y_array;
    y_array = E_CALLOC(cmodel->dim, sizeof(double));

    /* for every class */
    for (class_idx = 0; class_idx < cmodel->n_class; class_idx++) {
        size_t mn_attribute_count = 0;
        int *mn_attribute_indices = E_CALLOC(cmodel->dim, sizeof(int));

        /* for each attr in each class : classmodel7[i][l] */
        for (attr_idx = 0; attr_idx < cmodel->dim; attr_idx++) {

            switch (cmodel->classmodel7[class_idx][attr_idx]) {
                case SINGLE_NORMAL: {
                    double dist, factor;
                    if (test_vec[attr_idx] == BAD_DISTANCE_FLOAT) {
                        dist = 0;
                    } else {
                        dist = distance_to_mean(valid_distance_or(test_vec[attr_idx], 0),
                                                cmodel->params7[class_idx][attr_idx][MEAN],
                                                false);
                    }
                    double stddev = cmodel->params7[class_idx][attr_idx][STDDEV];
                    factor =
                            (gsl_sf_erf_Q((dist) / stddev - fabs(cmodel->abs_error_dist)) -
                             gsl_sf_erf_Q(((dist) / stddev + fabs(cmodel->abs_error_dist))));

                    if (factor != 0) {
                        mship[class_idx] *= factor;
                    } else {
                        mship[class_idx] *= 1;
                    }
                }
                    break;
                case MULTI_NORMAL: {
                    mn_attribute_indices[mn_attribute_count] = attr_idx;
                    if (test_vec[attr_idx] != BAD_ANGLE_FLOAT) {
                        y_array[mn_attribute_count] =
                                /* real location data */
                                distance_to_mean(valid_in_range_or(test_vec[attr_idx], -1, 1, 0),
                                                 cmodel->params7[class_idx][attr_idx][MEAN],
                                                 false);
                    } else {
                        y_array[mn_attribute_count] = 0;
                    }
                    mn_attribute_count++;
                }
                    break;
                case UNKNOWN:
                default:
                    err_printf(this_sub, "Unknown classmodel\n");
                    return NULL;
            }
        }

        if (mn_attribute_count > 0) {
            /****** initialize for MULTI_NORMAL Case *******/
            gsl_permutation *perm;
            gsl_matrix *C_inv;
            gsl_vector_view y;
            gsl_vector *C_inv_y;
            gsl_matrix *C = gsl_matrix_alloc(mn_attribute_count, mn_attribute_count);
            perm = gsl_permutation_alloc(mn_attribute_count);
            C_inv = gsl_matrix_alloc(mn_attribute_count, mn_attribute_count);
            C_inv_y = gsl_vector_alloc(mn_attribute_count);
            y = gsl_vector_view_array(y_array, mn_attribute_count);
            double exponent = 1.0;
            double C_det;
            int sign;
            size_t i, j;
            /**********  initialization finished  **********/

            for (i = 0; i < mn_attribute_count; i++) {
                int attr_i = mn_attribute_indices[i];
                for (j = 0; j < mn_attribute_count; j++) {
                    int attr_j = mn_attribute_indices[j];
                    double coeff = exp(cmodel->cov_matrix7[class_idx][(cmodel->dim * attr_i) + attr_j]);
                    gsl_matrix_set(C, i, j, coeff);
                }
            }

            // perform an LU decomposition of C (inplace)
            gsl_linalg_LU_decomp(C, perm, &sign);
            // and calculate the determinant of covariant matrix
            C_det = gsl_linalg_LU_det(C, sign);

            if (fabs(C_det) > 1e-30) {
                // calculate the inverse of the covariant matrix
                gsl_linalg_LU_invert(C, perm, C_inv);
                // calculate C_inv_y as: C_inv_y = C^-1 * y
                gsl_blas_dgemv(CblasNoTrans, 1.0, C_inv, &y.vector, 0.0, C_inv_y);
                // calculate (C^-1 * y) * y
                gsl_blas_ddot(C_inv_y, &y.vector, &exponent);

                double factor_mult = pow(2 * cmodel->abs_error_tau, mn_attribute_count) * /* not needed if normalised */
                                     exp(-0.5 * exponent) / sqrt(pow(2 * M_PI, mn_attribute_count) * fabs(C_det));

                if (factor_mult != 0.0) {
                    mship[class_idx] *= factor_mult;
                } else {
                    mship[class_idx] *= 1;
                }
            } else {
                mprintf("determinant is zero for class=%d\n", class_idx);
            }

            gsl_matrix_free(C);
            gsl_matrix_free(C_inv);
            gsl_vector_free(C_inv_y);
            gsl_permutation_free(perm);
            free_if_not_null(mn_attribute_indices);
        }
    }

    free_if_not_null(y_array);
    return mship;
}

#undef square

/* ------------------------ print_matrix ------------------------
 * help_function for printing a GSL matrix
*/
/*void print_matrix(gsl_matrix *ptr) {
    size_t i;
    for (i = 0; i < ptr->size1; i++) {
        size_t j;
        for (j = 0; j < ptr->size2; j++) {
            double value = gsl_matrix_get(ptr, i, j);
            mprintf("%1.4f ", (float) value);
        }
        mprintf("\n");
    }
    mprintf("\n");
}*/

float *
tau_computeMembership(float *mship, const float *test_vec,
                      const struct clssfcn7 *cmodel) {
    long double sum = 0.0;
    size_t i;
    for (i = 0; i < cmodel->n_class; i++) {
        mship[i] = cmodel->class_weight7[i];   /* Default class probability */
    }

    if (tau_computeMembershipStrct(mship, test_vec, cmodel) == NULL)
        return NULL;

    for (i = 0; i < cmodel->n_class; i++)
        sum += (mship[i] * mship[i]);
    sum = sqrt(sum);
    if (sum != 0.0)
        for (i = 0; i < cmodel->n_class; i++)
            mship[i] /= sum;
    return mship;
}

