// libWurstTest.cpp : Definiert den Einstiegspunkt fr die Konsolenanwendung.
//

#include "stdafx.h"

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

//#include "testLibWurst.h"
exter n "C" {
	#include "e_malloc.h"
	#include "mprintf.h"

	#include "yesno.h"
	#include "pair_set.h"

	#include "coord.h"
	#include "coord_i.h"
	#include "read_ac_strct_i.h"
	#include "align_i.h"
	#include "prob_vec_i.h"
	#include "score_mat_i.h"
	#include "read_seq_i.h"
	#include "score_probvec.h"
	#include "pair_set_i.h"
	#include "pair_set_p_i.h"
	#include "pair_set.h"
	#include "yesno.h"
	#include "pdbin_i.h"
	#include "pdbout_i.h"
	#include "lsqf.h"
}

struct score_struct{
	float scr_tot;
	float cvr;
};

const float sw1_pgap_open = 3.25;
const float sw1_pgap_widen = 0.8942;
const float mat_shift = -0.1;
/*   const int N_AND_W = 0;
*    const int S_AND_W = 1;
*    const char *filename = "/home/stud2004/tmargraf/andrew/scripts/classfile";
*    const char *pdbout = "/home/stud2010/ibondarenko/newpdb.pdb";
*/
/*-------------------------------- get_scores --------------------------------
 *
 * a port from salami's get_score function
 * from Steve Hoffmann etwas modifiziert
 *
 */


/* Auswertung vom Alignment (Abdeckung und Scor mit Gaps addition) */
static struct score_struct*
get_scores (struct pair_set *set, struct coord *a, struct coord *b)
{

    int cover=0,i ;
    float cover_float;
    char *pcover1;
    char *pcover2;
    struct seq* seq_a;
    struct score_struct *scores;
    float open_cost = 0, widen_cost = 0;
	
    scores = (struct score_struct*) E_MALLOC(sizeof(struct score_struct));

    pair_set_coverage(set, coord_size(a), coord_size(b),
                            &pcover1, &pcover2);
    for (i=0; i < coord_size(a); i++) {
        if(pcover1[i] == '1') cover++;
    }

    free(pcover1);
    free(pcover2);

    seq_a = coord_get_seq(a);
    cover_float = (float)cover/(float)seq_size(seq_a);

    if (cover_float >= 0.05) {
        pair_set_gap(set, &open_cost, &widen_cost, 1, 1);
    }

    scores->scr_tot = set->smpl_score + sw1_pgap_open*open_cost + sw1_pgap_widen*widen_cost;

    seq_destroy(seq_a);
    scores->cvr = cover_float;

    return scores;
}


/* ------------------  */

int _tmain(int argc, char* argv[])
{

    char *path1;
    char *path2;
    char *filename;
    char *pdbout;
    int alg_type;


    struct score_mat  *crap = NULL;
    struct pair_set *set_alg;
    struct score_struct *scores;
    float rmsd;
    FILE *classfile;


    /* read coordinates from pdb-datei.*/
    struct coord *coord1;
    struct coord *coord2;

    struct aa_strct_clssfcn *classstr;

    struct prob_vec *pvec1, *pvec2;
    struct seq *seq1, *seq2;

    struct score_mat  *matrix;
    int res;

    if (argc != 6) {
            mprintf("LIneformat prg <pdb1> <pdb2> <classFile> <pdbout> <alg type>\n");
            return EXIT_FAILURE;
    }

    path1= argv[1] ;
    path2 = argv[2];
    filename = argv[3];
    pdbout = argv[4];
    alg_type = atoi(argv[5]);

    /* read coordinates from pdb-datei.*/
    coord1 = pdb_read(path1, "", '_');
    coord2 = pdb_read(path2, "", '_');


    /* read the classification structur */
    classfile = fopen(filename, "r");
    if (classfile == NULL) {
        mprintf("The classfile is not found. \n");
        return EXIT_FAILURE;
    }
    fclose(classfile);

    classstr = aa_strct_clssfcn_read (filename, 0.4);

    pvec1 = strct_2_prob_vec(coord1, classstr, YES );
    pvec2 = strct_2_prob_vec(coord2, classstr, YES );
    seq1 = coord_get_seq(coord1);
    seq2 = coord_get_seq(coord2);

    /* ausrechnung der Matrix */
    matrix = score_mat_new(seq_size(seq1), seq_size(seq2));
    matrix = score_mat_shift(matrix, mat_shift);
    res = score_pvec(matrix, pvec1, pvec2);

    /* Alignment (den Pfad) berechnung	 */
    set_alg = score_mat_sum_full(&crap, matrix,
                             sw1_pgap_open, sw1_pgap_widen,
                             sw1_pgap_open, sw1_pgap_widen,
                             NULL, NULL, alg_type, NULL);

    /* Auswertung vom Alignment */
    scores = get_scores(set_alg, coord1, coord2);

    if (coord_rmsd(set_alg, coord1, coord2, 0, &rmsd, &coord1, &coord2) == EXIT_SUCCESS) {
        coord_2_pdb(pdbout, coord2, seq2);
    }
    else {
        mprintf("The coord_rmsd was not successful. \n");
    }

    mprintf ("score_pvec = %d auswertung scores = %f \n", res, scores->scr_tot);
    mprintf ("S1: %s \n", seq_print(seq1));
    pair_set_extend(set_alg, seq_size(seq1), seq_size(seq2), EXT_LONG);
    mprintf ("%s \n", pair_set_pretty_string(set_alg, seq1, seq2, NULL, NULL));
    mprintf ("S2: %s \n", seq_print(seq2));
    score_mat_destroy(crap);
    pair_set_destroy(set_alg);
    free(scores);

    score_mat_destroy(matrix);
    prob_vec_destroy(pvec1);
    prob_vec_destroy(pvec2);

    return 0;
}

